%{
#include "node.h"
#include "parser.h"

int yyerror(char*);

int num_lines = 1;

%}

%option noyywrap

%%

"--".* {/* ignore */}

\n {++num_lines;}

[ \t]+ {/* ignore */}

"system" {
yylval.lineno = num_lines;
return(SYSTEM);
}

"implementation" {
yylval.lineno = num_lines;
return(IMPLEMENTATION);
}

"end" {
yylval.lineno = num_lines;
return(END);
}

":" {
yylval.lineno = num_lines;
return(COLON);
}

";" {
yylval.lineno = num_lines;
return(SEMICOLON);
}

"." {
yylval.lineno = num_lines;
return(DOT);
}

"bool" {
yylval.lineno = num_lines;
return(BOOLEAN);
}

"int" {
yylval.lineno = num_lines;
return(INTEGER);
}

"index" {
yylval.lineno = num_lines;
return(INTEGER);
}

"parameters" {
yylval.lineno = num_lines;
return(PARAMETERS);
}

"set" {
yylval.lineno = num_lines;
return(SET);
}

"+" {
yylval.lineno = num_lines;
return(PLUS);
}

"-" {
yylval.lineno = num_lines;
return(MINUS);
}

"*" {
yylval.lineno = num_lines;
return(TIMES);
}

"next" {
yylval.lineno = num_lines;
return(NEXT);
}

"of"  {
yylval.lineno = num_lines;
return(OF);
}

"indices"  {
yylval.lineno = num_lines;
return(INDICES);
}

"indexed"  {
yylval.lineno = num_lines;
return(INDEXED);
}

"by"  {
yylval.lineno = num_lines;
return(BY);
}

"true"|"BOOL_TRUE" {
yylval.node = new_node(BOOL_TRUE, Nil, Nil);
return(BOOL_TRUE);
}

"false"|"BOOL_FALSE" {
yylval.node = new_node(BOOL_FALSE, Nil, Nil);
return(BOOL_FALSE);
}

","  {
yylval.lineno = num_lines;
return(COMMA);
}

"invariants"  {
yylval.lineno = num_lines;
return(INVARIANTS);
}

"initial"  {
yylval.lineno = num_lines;
return(INIT);
}

"subcomponents"  {
yylval.lineno = num_lines;
return(SUBCOMPONENTS);
}

"single"  {
yylval.lineno = num_lines;
return(SINGLE);
}

"connections"  {
yylval.lineno = num_lines;
return(CONNECTIONS);
}

"port"  {
yylval.lineno = num_lines;
return(PORT);
}

"if"  {
yylval.lineno = num_lines;
return(IF);
}

"for"  {
yylval.lineno = num_lines;
return(FOR);
}

"in"  {
yylval.lineno = num_lines;
return(IN);
}

"reconfigurations"  {
yylval.lineno = num_lines;
return(RECONFIGURATIONS);
}

"size"  {
yylval.lineno = num_lines;
return(SIZE);
}

"add"  {
yylval.lineno = num_lines;
return(ADD);
}

"remove"  {
yylval.lineno = num_lines;
return(REMOVE);
}

"exists"  {
yylval.lineno = num_lines;
return(EXISTS);
}

"forall"  {
yylval.lineno = num_lines;
return(FORALL);
}

"!"|"not"  {
yylval.lineno = num_lines;
return(NOT);
}

"->"  {
yylval.lineno = num_lines;
return(IMPLIES);
}

"<->"  {
yylval.lineno = num_lines;
return(IFF);
}

"&"|"and"  {
yylval.lineno = num_lines;
return(AND);
}

"|"|"or" {
yylval.lineno = num_lines;
return(OR);
}

"["  {
yylval.lineno = num_lines;
return(LB);
}

"]"  {
yylval.lineno = num_lines;
return(RB);
}

"{"  {
yylval.lineno = num_lines;
return(LC);
}

"}"  {
yylval.lineno = num_lines;
return(RC);
}

"("  {
yylval.lineno = num_lines;
return(LP);
}

")"  {
yylval.lineno = num_lines;
return(RP);
}

"/"  {
yylval.lineno = num_lines;
return(FRAC);
}

"<="  {
yylval.lineno = num_lines;
return(LE);
}

">="  {
yylval.lineno = num_lines;
return(GE);
}

"="  {
yylval.lineno = num_lines;
return(EQUAL);
}

"<"  {
yylval.lineno = num_lines;
return(LT);
}

">"  {
yylval.lineno = num_lines;
return(GT);
}

"!="  {
yylval.lineno = num_lines;
return(NOTEQUAL);
}

[-]?[0-9]+ {
long tmp;
char* endptr;

errno = 0;
tmp = strtol(yytext, &endptr, 10);

/* range errors */
if ((errno == ERANGE)
    || (errno != 0 && tmp == 0)

    /* no digits found */
    || (endptr == yytext)

    /* garbage after digits */
    || (*endptr != '\0')) {

fprintf(stderr, "illegal integer: \"%s\"", yytext);
} else {
yylval.node = new_number_node(tmp);
return(NUMBER);
}
}

[A-Za-z_][A-Za-z0-9_]* {
char* new_prop = malloc(strlen(yytext)+1);
strcpy(new_prop, yytext);
yylval.node = new_node(ID, (node_ptr)new_prop, Nil);
return(ID);
}

. {
char msg[25];
sprintf(msg,"%s <%s>","invalid character",yytext);
yyerror(msg);
}
