#include "node.h"
#include "parser.h"
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

struct node_TAG {
  int type;
  node_ptr left;
  node_ptr right;
  int number;
};

typedef struct node_TAG node;

node_ptr new_node(int type, node_ptr left, node_ptr right) {
  node_ptr n = (node_ptr)malloc(sizeof(node));
  n->type = type;
  n->left = left;
  n->right = right;
  return n;
}

int node_get_type(node_ptr n) {
  assert(NULL!=n);
  return n->type;
}

node_ptr *node_get_left_ptr(node_ptr n) {
  assert(NULL!=n);
  return &(n->left);
}

node_ptr *node_get_right_ptr(node_ptr n) {
  assert(NULL!=n);
  return &(n->right);
}

node_ptr node_get_left(node_ptr n) {
  assert(NULL!=n);
  return n->left;
}

node_ptr node_get_right(node_ptr n) {
  assert(NULL!=n);
  return n->right;
}

int node_get_number(node_ptr n) {
  assert(NULL!=n);
  return n->number;
}

node_ptr new_number_node(int number) {
  node_ptr n = (node_ptr)malloc(sizeof(node));
  n->type = NUMBER;
  n->left = Nil;
  n->right = Nil;
  n->number = number;
  return n;
}

void free_node(node_ptr n){
  if (NULL!=n) {
    free(n->left);
    free(n->right);
    free(n);
  }
}

/* WARNING: this function is still a draft */
void print_node(node_ptr n, FILE* out) {
  if (NULL==n) return;
  switch (n->type) {
  case CONS:
    print_node(n->left, out);
    if (NULL!=n->right)
      print_node(n->right, out);
    break;
  case PARAMETERS: {
    fprintf(out, "parameters\n");
    print_node(n->left, out);
    break;
  }
  case INVARIANTS: {
    fprintf(out, "invariants\n");
    print_node(n->left, out);
    break;
  }
  case SUBCOMPONENTS: {
    fprintf(out, "subcomponents\n");
    print_node(n->left, out);
    break;
  }
  case PARAMETER: {
    print_node(n->left, out);
    fprintf(out, ": ");
    print_node(n->right, out);
    fprintf(out, ";\n");
    break;
  }
  case SUBCOMPONENT: {
    print_node(n->left->left, out);
    fprintf(out, ": ");
    if (SINGLE==n->left->right->type)
      fprintf(out, "single");
    else {
      fprintf(out, "set indexed by ");
      print_node(n->left->right, out);
    }
    print_node(n->right, out);
    fprintf(out, ";\n");
    break;
  }
  case GUARD: {
    if (BOOL_TRUE!=n->left->type) {
      fprintf(out, "if ");
      print_node(n->left, out);
    }
    break;
  }
  case INDEX_SET_TYPE:
    fprintf(out, "set of indices");
    break;
  case ID: {
    char *prop_name = (char *)n->left;
    fprintf(out, "%s", prop_name);
    break;
  }
  case BOOLEAN:
    fprintf(out, "boolean");
    break;
  case INTEGER:
    fprintf(out, "integer");
    break;
  case BOOL_TRUE:
    fprintf(out, "BOOL_TRUE");
    break;
  case BOOL_FALSE:
    fprintf(out, "BOOL_FALSE");
    break;
  case MINUS:
    fprintf(out, "( ");
    print_node(n->left, out);
    fprintf(out, " - ");
    print_node(n->right, out);
    fprintf(out, " )");
    break;
  case PLUS:
    fprintf(out, "( ");
    print_node(n->left, out);
    fprintf(out, " + ");
    print_node(n->right, out);
    fprintf(out, " )");
    break;
  case TIMES:
    fprintf(out, "( ");
    print_node(n->left, out);
    fprintf(out, " * ");
    print_node(n->right, out);
    fprintf(out, " )");
    break;
  case NEXT:
    fprintf(out, "next(");
    print_node(n->left, out);
    fprintf(out, ")");
    break;
  case ITE:
    fprintf(out, "( ");
    print_node(n->left, out);
    fprintf(out, " ? ");
    print_node(n->right->left, out);
    fprintf(out, " : ");
    print_node(n->right->right, out);
    fprintf(out, " )");
    break;
  case AND:
    fprintf(out, "( ");
    print_node(n->left, out);
    fprintf(out, " & ");
    print_node(n->right, out);
    fprintf(out, " )");
    break;
  case OR:
    fprintf(out, "( ");
    print_node(n->left, out);
    fprintf(out, " | ");
    print_node(n->right, out);
    fprintf(out, " )");
    break;
  case IMPLIES:
    fprintf(out, "( ");
    print_node(n->left, out);
    fprintf(out, " -> ");
    print_node(n->right, out);
    fprintf(out, " )");
    break;
  case IFF:
    fprintf(out, "( ");
    print_node(n->left, out);
    fprintf(out, " <-> ");
    print_node(n->right, out);
    fprintf(out, " )");
    break;
  case NOT:
    fprintf(out, "( ! ");
    print_node(n->left, out);
    fprintf(out, ") ");
    break;
  case NUMBER: {
    fprintf(out, "%d", n->number);
    break;
  }
  case GT:
    fprintf(out, "(");
    print_node(n->left, out);
    fprintf(out, ">");
    print_node(n->right, out);
    fprintf(out, ")");
    break;
  case GE:
    fprintf(out, "(");
    print_node(n->left, out);
    fprintf(out, ">=");
    print_node(n->right, out);
    fprintf(out, ")");
    break;
  case LT:
    fprintf(out, "(");
    print_node(n->left, out);
    fprintf(out, "<");
    print_node(n->right, out);
    fprintf(out, ")");
    break;
  case LE:
    fprintf(out, "(");
    print_node(n->left, out);
    fprintf(out, "<=");
    print_node(n->right, out);
    fprintf(out, ")");
    break;
  case EQUAL:
    fprintf(out, "(");
    print_node(n->left, out);
    fprintf(out, "=");
    print_node(n->right, out);
    fprintf(out, ")");
    break;
  case NOTEQUAL:
    fprintf(out, "(");
    print_node(n->left, out);
    fprintf(out, "!=");
    print_node(n->right, out);
    fprintf(out, ")");
    break;
  default:
    fprintf(stderr, "Wrong expression in node_print of type %d\n", n->type);
  }
}

/* return 0 iff n1 and n2 are different
*/
int node_cmp(node_ptr n1, node_ptr n2){
  if (n1==NULL || n2==NULL)
    return n1==n2;
  if (n1->type!=n2->type)
    return 0;
  else
    switch (n1->type) {
      /* n2 has the same type */
    case ID:
      return 0==strcmp((char*)n1->left, (char*)n2->left);
    default:
      if (0==node_cmp(n1->left,n2->left))
        return 0;
      else
        return node_cmp(n1->right,n2->right);
    }
  fprintf(stderr, "Unreachable code in node_cmp: found nodes\n");
  print_node(n1, stderr);
  fprintf(stderr, "\n");
  print_node(n2, stderr);
  fprintf(stderr, "\n");
  assert(0);
  return 1;
}
