%{
#include "node.h"
#include "in2mcmt.h"
#include <stdio.h>
#include <stdlib.h>
#include <glib.h>

#define SRC_OPTION "-src"
#define DST_OPTION "-dst"


  extern int num_lines;
  extern char* yytext;

  int yylex(void);
  int yyerror(char*);

  node_ptr parse_tree;

  GHashTable *src;
  GHashTable *dst;

  void parse_args(int, char **);
  void print_usage(char *);
%}


%union {
  node_ptr node;
  int lineno;
}

%token <node> NUMBER
%token <node> ID
%token <node> BOOL_TRUE BOOL_FALSE
%token <lineno> DOT CONS
%token <lineno> BOOLEAN REAL INTEGER
%token <lineno> SYSTEM IMPLEMENTATION END PARAMETERS INVARIANTS INIT SUBCOMPONENTS SINGLE RECONFIGURATIONS
%token <lineno> PARAMETER SUBCOMPONENT CONNECTION PAIR GUARD SCOPE SCOPE_COMPLEMENT
%token <lineno> SET OF INDICES INDEXED BY CONNECTIONS IF FOR ITERATOR PORT
%token <lineno> INDEX_SET_TYPE SET_TYPE
%token <lineno> IN ADD REMOVE SIZE EXISTS FORALL
%token <lineno> NOT AND OR IMPLIES IFF ITE
%token <lineno> LB RB LP RP LC RC
%token <lineno> GE LE EQUAL GT LT NOTEQUAL
%token <lineno> PLUS MINUS TIMES FRAC
%token <lineno> COLON SEMICOLON COMMA
%token <lineno> NEXT

%left OR IMPLIES
%left AND
%left IFF
%precedence NOT
%left EQUAL NOTEQUAL LT LE GT GE
%left PLUS MINUS
%left TIMES FRAC

%type <node> system
%type <node> declaration_list
%type <node> declaration
%type <node> parameters
%type <node> parameter
%type <node> parameter_list
%type <node> id_list
%type <node> index
%type <node> index_list
%type <node> iter_id_list
%type <node> indexed_identifier
%type <node> invariants
%type <node> initial
%type <node> subcomponents
%type <node> subcomponent
%type <node> subcomponent_list
%type <node> subcomponent_type
%type <node> connections
%type <node> connection
%type <node> connection_list
%type <node> subcomponent_spec
%type <node> access_list
%type <node> guard
%type <node> iterator
%type <node> reconfigurations
%type <node> transition_expression
%type <node> transition_expression_list
%type <node> boolean_constant
%type <node> integer_constant
%type <node> expression
%type <node> expression_list
%type <node> to_add_rem
%type <node> type
%type <node> simple_type
%type <node> term

%start system
%%

system : SYSTEM IMPLEMENTATION ID DOT ID declaration_list END ID DOT ID SEMICOLON
{
  parse_tree = $6;
} |
SYSTEM declaration_list END SEMICOLON
{
  parse_tree = $2;
};

declaration_list: declaration
{
  $$ = new_node(CONS, $1, NULL);
} |
declaration_list declaration
{
  $$ = new_node(CONS, $2, $1);
};

declaration:
parameters
| invariants
| initial
| subcomponents
| connections
| reconfigurations;

parameters: PARAMETERS parameter_list
{
  $$ = new_node(PARAMETERS, $2, NULL);
};

parameter_list: parameter
{
  $$ = new_node(CONS, $1, NULL);
} |
parameter_list parameter
{
  $$ = new_node(CONS, $2, $1);
};

parameter: ID COLON type SEMICOLON
{
  $$ = new_node(PARAMETER, $1, $3);
};

simple_type: INTEGER
{
  $$ = new_node(INTEGER, NULL, NULL);
} |
BOOLEAN
{
  $$ = new_node(BOOLEAN, NULL, NULL);
};

type: SET OF INDICES
{
  $$ = new_node(INDEX_SET_TYPE, NULL, NULL);
} |
simple_type |
SET INDEXED BY ID OF simple_type
{
  $$ = new_node(SET_TYPE,
                new_node(PAIR, $4, $6), NULL);
} |
SET INDEXED BY LP id_list RP OF simple_type
{
  $$ = new_node(SET_TYPE,
                new_node(PAIR, $5, $8), NULL);
};

boolean_constant: BOOL_TRUE |
BOOL_FALSE ;

integer_constant: NUMBER ;

id_list: ID |
id_list COMMA ID
{
  $$ = new_node(CONS, $3, $1);
};

iter_id_list: ID IN ID
{
  $$ = new_node(CONS, new_node(IN, $1, $3), NULL);
} |
iter_id_list COMMA ID IN ID
{
  $$ = new_node(CONS, new_node(IN, $3, $5), $1);
};

invariants: %empty
{
  $$ = new_node(INVARIANTS, NULL, NULL);
} |
INVARIANTS expression_list
{
  $$ = new_node(INVARIANTS, $2, NULL);
};

expression_list: expression SEMICOLON |
expression_list expression SEMICOLON
{
  $$ = new_node(CONS, $2, $1);
};

initial: INIT expression SEMICOLON
{
  $$ = new_node(INIT, $2, NULL);
};

subcomponents: SUBCOMPONENTS subcomponent_list
{
  $$ = new_node(SUBCOMPONENTS, $2, NULL);
};

subcomponent_list: subcomponent
{
  $$ = new_node(CONS, $1, NULL);
} |
subcomponent_list subcomponent
{
  $$ = new_node(CONS, $2, $1);
};

subcomponent: ID COLON subcomponent_type guard SEMICOLON
{
  $$ = new_node(SUBCOMPONENT,
                new_node(PAIR, $1, $3),
                $4);
};

subcomponent_type: SINGLE
{
  $$ = new_node(SINGLE, NULL, NULL);
} |
SET INDEXED BY ID
{
  $$ = $4;
} |
SYSTEM ID
{
  $$ = new_node(SINGLE, NULL, NULL);
} |
SET INDEXED BY ID OF SYSTEM ID
{
  $$ = $4;
};

connections: CONNECTIONS connection_list
{
  $$ = new_node(CONNECTIONS, $2, NULL);
};

connection_list: connection |
connection_list connection
{
  $$ = new_node(CONS, $2, $1);
};

connection: subcomponent_spec IMPLIES subcomponent_spec guard iterator SEMICOLON
{
  $$ = new_node(CONNECTION,
                new_node(PAIR, $1, $3),
                new_node(PAIR, $4, $5));
} |
ID COLON PORT subcomponent_spec IMPLIES subcomponent_spec guard iterator SEMICOLON
{
  $$ = new_node(CONNECTION,
                new_node(PAIR, $4, $6),
                new_node(PAIR, $7, $8));
};

subcomponent_spec: ID |
indexed_identifier |
ID DOT ID { $$ = $1; } |
indexed_identifier DOT ID { $$ = $1; };

indexed_identifier: ID access_list
{ $$ = new_node(INDEXED, $1, $2); };

access_list: LB expression RB
{ $$ = new_node(CONS, $2, NULL); } |
access_list LB expression RB
{ $$ = new_node(CONS, $3, $1); };

guard: %empty
{
  $$ = new_node(GUARD, new_node(BOOL_TRUE, NULL, NULL), NULL);
} |
IF expression
{
  $$ = new_node(GUARD, $2, NULL);
};

iterator: %empty
{
  $$ = new_node(ITERATOR, NULL, NULL);
} |
FOR iter_id_list
{
  $$ = new_node(ITERATOR, $2, NULL);
};

reconfigurations: RECONFIGURATIONS transition_expression_list
{
  $$ = new_node(RECONFIGURATIONS, $2, NULL);
};

transition_expression_list: transition_expression SEMICOLON
{
  $$ = new_node(CONS, $1, NULL);
} |
transition_expression_list transition_expression SEMICOLON
{
  $$ = new_node(CONS, $2, $1);
};

transition_expression: expression;

expression :
term |
expression  EQUAL expression { $$ = new_node(EQUAL, $1, $3); } |
expression  NOTEQUAL expression { $$ = new_node(NOTEQUAL, $1, $3); } |
expression  LE expression { $$ = new_node(LE, $1, $3); } |
expression  GE expression { $$ = new_node(GE, $1, $3); } |
expression  LT expression { $$ = new_node(LT, $1, $3); } |
expression  GT expression { $$ = new_node(GT, $1, $3); } |
NOT expression
{ $$ = new_node(NOT, $2, NULL); } |
expression PLUS expression
{ $$ = new_node(PLUS, $1, $3); } |
expression MINUS expression
{ $$ = new_node(MINUS, $1, $3); } |
expression TIMES expression
{ $$ = new_node(TIMES, $1, $3); } |
expression FRAC expression
{ $$ = new_node(FRAC, $1, $3); } |
expression IMPLIES expression
{ $$ = new_node(IMPLIES, $1, $3); } |
expression IFF expression
{ $$ = new_node(IFF, $1, $3); } |
expression AND expression
{ $$ = new_node(AND, $1, $3); } |
expression OR expression
{ $$ = new_node(OR, $1, $3); } |
term IN term
{ $$ = new_node(IN, $1, $3); } |
term NOT IN term
{ $$ = new_node(NOT,
                new_node(IN, $1, $4),
                NULL); } |
LP expression RP
{ $$ = $2; } |
EXISTS LP ID IN ID COMMA expression RP
{ $$ = new_node(EXISTS,
                new_node(SCOPE, $3, $5),
                $7);
} |
EXISTS LP ID NOT IN ID COMMA expression RP
{ $$ = new_node(EXISTS,
                new_node(SCOPE_COMPLEMENT, $3, $6),
                $8);
} |
FORALL LP ID IN ID COMMA expression RP
{ $$ = new_node(FORALL,
                new_node(SCOPE, $3, $5),
                $7);
} |
FORALL LP ID NOT IN ID COMMA expression RP
{ $$ = new_node(FORALL,
                new_node(SCOPE_COMPLEMENT, $3, $6),
                $8);
};

term: ID |
indexed_identifier |
boolean_constant |
integer_constant |
LC index_list RC
{ $$ = $2; } |
ADD LP ID COMMA to_add_rem RP
{ $$ = new_node(ADD, $3, $5); } |
REMOVE LP ID COMMA to_add_rem RP
{ $$ = new_node(REMOVE, $3, $5); } |
SIZE LP ID RP
{ $$ = new_node(SIZE, $3, NULL); } |
NEXT LP ID RP
{ $$ = new_node(NEXT, $3, NULL); } |
NEXT LP indexed_identifier RP
{ $$ = new_node(NEXT, $3, NULL); } ;

to_add_rem: index |
LC index_list RC
{ $$ = $2; };

index_list: %empty
{ $$ = NULL; } |
index
{ $$ = new_node(CONS, $1, NULL); } |
index_list COMMA index
{ $$ = new_node(CONS, $3, $1); } ;

index: ID | NUMBER;

%%

int yyerror(char *s) {
  fprintf(stderr, "%s line: %d, token: %s\n", s, num_lines, yytext);
  return 0;
}

int main(int argc, char **argv) {
  src = g_hash_table_new(g_str_hash, g_str_equal);
  dst = g_hash_table_new(g_str_hash, g_str_equal);

  yyparse();

  if (argc>1) {
    parse_args(argc, argv);
    input_to_mcmt(parse_tree, src, dst);
  }

  g_hash_table_destroy(src);
  g_hash_table_destroy(dst);

  return 0;
}

void parse_args(int argc, char **argv) {
    int i;
    int in_src = 1;

    if (argc < 2 || strcmp(argv[1], SRC_OPTION)) {
        print_usage(argv[0]);
        exit(1);
    }

    for (i = 2; i < argc; i++) {
        if (!strcmp(argv[i], DST_OPTION)) {
            if (g_hash_table_size(src) == 0) {
                fprintf(stderr, "must specify at least one source\n");
                exit(1);
            }
            in_src = 0;
        } else if (in_src) {
            g_hash_table_add(src, argv[i]);
        } else {
            g_hash_table_add(dst, argv[i]);
        }
    }

    if (in_src) {
        print_usage(argv[0]);
        exit(1);
    }

    if (g_hash_table_size(dst) == 0) {
        fprintf(stderr, "must specify at least one destination\n");
        exit(1);
    }
}

void print_usage(char *name) {
    fprintf(stderr, "Usage: %s %s s_1 ... s_m %s d_1 ... d_n\n",
            name, SRC_OPTION, DST_OPTION);
}
