indent = 0

def generate(size, outfile):
    with open(outfile, "w") as f:
        gen_init(size, f)
        gen_parameters(size, f)
        gen_invariants(size, f)
        gen_initial(size, f)
        gen_subcomponents(size, f)
        gen_connections(size, f)
        gen_reconfigurations(size, f)
        gen_end(size, f)

def get_src_dst(size):
	return ("a0", "a"+str(size))

def gen_init(sz, f):
    write(f, "system implementation sys.impl\n")

def gen_parameters(sz, f):
    global indent
    indent += 1
    write(f, "parameters\n")
    indent += 1
    for i in range(0, sz + 1):
        write(f, "A"+str(i)+": set of indices;\n")
    for i in range(0, sz):
        write(f, "m"+str(i)+"_conn_a"+str(i)+": int;\n")
        write(f, "m"+str(i)+"_conn_a"+str(i+1)+": int;\n")
        write(f, "m"+str(i)+"_present: bool;\n")
    indent -= 2
    write(f, "\n")

def gen_invariants(sz, f):
    pass

def gen_initial(sz, f):
    global indent
    indent += 1
    write(f, "initial\n")
    indent += 1
    write(f, "")
    for i in range(0, sz):
        f.write("m"+str(i)+"_conn_a"+str(i)+" = 0 and ")
        f.write("m"+str(i)+"_conn_a"+str(i+1)+" = 0")
        if i < sz - 1:
            f.write(" and ")
    f.write(" and ")
    for i in range(0, sz+1):
        f.write("forall(x in A"+str(i)+", x > 0)")
        if i < sz:
            f.write(" and ")
    f.write(";\n")
    indent -= 2
    write(f, "\n")

def gen_subcomponents(sz, f):
    global indent
    indent += 1
    write(f, "subcomponents\n")
    indent += 1
    for i in range(0, sz + 1):
        write(f, "a"+str(i)+": set indexed by A"+str(i)+";\n")
    for i in range(0, sz):
        write(f, "m"+str(i)+": single if m"+str(i)+"_present;\n")
    indent -= 2
    write(f, "\n")

def gen_connections(sz, f):
    global indent
    indent += 1
    write(f, "connections\n")
    indent += 1
    for i in range(0, sz):
        write(f, "a"+str(i)+"[x] -> m"+str(i)+" if x = m"+str(i)+"_conn_a"
                 +str(i)+" for x in A"+str(i)+";\n")
        write(f, "m"+str(i)+" -> a"+str(i)+"[x] if x = m"+str(i)+"_conn_a"
                 +str(i)+" for x in A"+str(i)+";\n")
        write(f, "a"+str(i+1)+"[x] -> m"+str(i)+" if x = m"+str(i)+"_conn_a"
                 +str(i+1)+" for x in A"+str(i+1)+";\n")
        write(f, "m"+str(i)+" -> a"+str(i+1)+"[x] if x = m"+str(i)+"_conn_a"
                 +str(i+1)+" for x in A"+str(i+1)+";\n")
    indent -= 2
    write(f, "\n")

def gen_reconfigurations(sz, f):
    global indent
    indent += 1
    write(f, "reconfigurations\n")
    indent += 1
    for i in range(0, sz+1):
        write(f, "exists(x not in A"+str(i)+", x > 0 and next(A"+str(i)
                 +") = add(A"+str(i)+", x));\n")
    for i in range(0, sz+1):
        write(f, "exists(x in A"+str(i)+", next(A"+str(i)+") = remove(A"
                 +str(i)+", x)")
        if i > 0:
            f.write(" and m"+str(i-1)+"_conn_a"+str(i)+" != x")
        if i < sz:
            f.write(" and m"+str(i)+"_conn_a"+str(i)+" != x")
        f.write(");\n")
    for i in range(0, sz):
        write(f, "exists(x in A"+str(i)+", m"+str(i)+"_conn_a"+str(i+1)
                 +" = 0 and m"+str(i)+"_conn_a"+str(i)+" = 0 and "
                 +"next(m"+str(i)+"_conn_a"+str(i)+") = x);\n")
        write(f, "exists(x in A"+str(i+1)+", m"+str(i)+"_conn_a"+str(i)
                 +" = 0 and m"+str(i)+"_conn_a"+str(i+1)+" = 0 and "
                 +"next(m"+str(i)+"_conn_a"+str(i+1)+") = x);\n")
    for i in range(0, sz):
        write(f, "next(m"+str(i)+"_conn_a"+str(i)+") = 0;\n")
        write(f, "next(m"+str(i)+"_conn_a"+str(i+1)+") = 0;\n")
    for i in range(0, sz):
        write(f, "m"+str(i)+"_conn_a"+str(i)+" = 0 and m"+str(i)+"_conn_a"
                 +str(i+1)+" = 0 and next(m"+str(i)
                 +"_present) = false;\n")
    for i in range(0, sz):
        write(f, "next(m"+str(i)+"_present) = true;\n")
    indent -= 2

def gen_end(sz, f):
    write(f, "end sys.impl;\n")

def write(f, s):
    global indent
    for i in range(0, indent):
        f.write("    ")
    f.write(s)
