# scalable network example with layers (unsafe version)
# d -> s -> c0 -> c1 -> ... -> c_{size-1} -> e
# \__________/ |  \__________________________/
#     |        |             all evil
#     |        |
#     |       this connection possible only if
#     |       c0 is untrusted
#     |
# d, s, c0 work like in
# the original example

indent = 0

def generate(size, outfile):
    with open(outfile, "w") as f:
        gen_init(size, f)
        gen_parameters(size, f)
        gen_invariants(size, f)
        gen_initial(size, f)
        gen_subcomponents(size, f)
        gen_connections(size, f)
        gen_reconfigurations(size, f)
        gen_end(size, f)

def get_src_dst(size):
	return ("d", "e")

def gen_init(sz, f):
    write(f, "system implementation sys.impl\n")

def gen_parameters(sz, f):
    global indent
    indent += 1
    write(f, "parameters\n")
    indent += 1
    for i in range(0, sz):
        write(f, "C"+str(i)+": set of indices;\n")
        if i == 0:
            write(f, "trustedClients"+str(i)+": set indexed by "
                    +"C"+str(i)+" of bool;\n")
        write(f, "connectedTo"+str(i)+": set indexed by "
                +"C"+str(i)+" of index;\n")
    write(f, "S: set of indices;\n")
    write(f, "trustedServers: set indexed by S of bool;\n")
    indent -= 2
    write(f, "\n")

def gen_invariants(sz, f):
    pass

def gen_initial(sz, f):
    global indent
    indent += 1
    write(f, "initial\n")
    indent += 1
    write(f, ("forall(c in C0, forall (s in S, (not trustedClients0[c] "
              "and s = connectedTo0[c]) -> (not trustedServers[s])))\n"))
    if sz > 1:
        write(f, ("and forall(c in C1, forall (d in C0, d = connectedTo1[c] "
                  "-> (not trustedClients0[d])))\n"))
    write(f, "and forall(c in C0, forall (s not in S, s != connectedTo0[c]))")
    if sz > 1:
        f.write("\n")

    for i in range(0, sz-1):
        write(f, "and forall(c in C"+str(i+1)+", forall (d not in C"+str(i)
                 +", d != connectedTo"+str(i+1)+"[c]))")
        if i < sz-2:
            f.write("\n")
    f.write(";")

    f.write("\n")
    indent -= 2
    write(f, "\n")

def gen_subcomponents(sz, f):
    global indent
    indent += 1
    write(f, "subcomponents\n")
    indent += 1
    write(f, "d: system databaseServer;\n")
    write(f, "servers: set indexed by S of system applicationServer;\n")
    for i in range(0, sz):
        write(f, "clients"+str(i)+": set indexed by C"+str(i)
                 +" of system Client;\n")
    write(f, "e: system Evil;\n")
    indent -= 2
    write(f, "\n")

def gen_connections(sz, f):
    global indent
    indent += 1
    write(f, "connections\n")
    indent += 1
    write(f, "d -> servers[s] if"
             +" trustedServers[s] for s in S;\n")
    write(f, "servers[s] -> clients0[c]"
             +" if s = connectedTo0[c] for s in S, c in C0;\n")
    for i in range(0, sz-1):
        write(f, "clients"+str(i)+"[c] -> clients"+str(i+1)+"[d]"
                 +" if c = connectedTo"+str(i+1)+"[d] for c in C"+str(i)
                 +", d in C"+str(i+1)+";\n")
    for i in range(0, sz):
        write(f, "clients"+str(i)+"[c] -> clients"+str(i)+"[d]")
        if i == 0:
            f.write(" if not trustedClients"+str(i)+"[c]")
        f.write(" for c in C"+str(i)
                 +", d in C"+str(i)+";\n")
    if sz == 1:
        write(f, "clients0[c] -> e if not trustedClients0[c] for c in C0;\n")
    else:
        write(f, "clients"+str(sz-1)+"[c] -> e for c in C"+str(sz-1)+";\n")
    indent -= 2
    write(f, "\n")

def gen_reconfigurations(sz, f):
    global indent
    indent += 1
    write(f, "reconfigurations\n")
    indent += 1
    write(f, ("exists(s not in S, next(S) = add(S, s) "
              "and next(trustedServers[s]) = true);\n"))
    write(f, ("exists(s in S, "
              "next(trustedServers[s]) = false);\n"))
    f.write("\n")

    write(f, ("exists(s in S, exists (c not in C0, not trustedServers[s] "
              "and next(C0) = add(C0, c) and next(connectedTo0[c]) = s "
              "and next(trustedClients0[c]) = false));\n"))
    for i in range(0, sz-1):
        write(f, "exists(c in C"+str(i)+", exists (d not in C"+str(i+1)+",")
        if i == 0:
            f.write(" not trustedClients"+str(i)+"[c] and")
        f.write(" next(C"+str(i+1)+") = add(C"+str(i+1)
                +", d) and next(connectedTo"+str(i+1)+"[d]) = c));\n")
    f.write("\n")

    write(f, ("exists(s in S, exists (c not in C0, trustedServers[s] "
              "and next(C0) = add(C0, c) and next(connectedTo0[c]) = s "
              "and next(trustedClients0[c]) = true));\n"))
    f.write("\n")

    write(f, ("exists(s in S, exists (c in C0, trustedClients0[c] and "
              "next(connectedTo0[c]) = s));\n"))
    f.write("\n")

    write(f, ("exists(s in S, exists (c in C0, not trustedClients0[c] and "
              "not trustedServers[s] and "
              "next(connectedTo0[c]) = s));\n"))
    for i in range(0, sz-1):
        write(f, "exists(c in C"+str(i)+", exists (d in C"+str(i+1) +",")
        if i == 0:
            f.write(" not trustedClients"+str(i)+"[c] and")
        f.write(" next(connectedTo"+str(i+1)+"[d]) = c));\n")
    f.write("\n")

    write(f, ("exists(s in S, exists (c in C0, not trustedClients0[c] "
              "and next(connectedTo0[c]) = s "
              "and next(trustedServers[s]) = false));\n"))
    f.write("\n")

    write(f, ("exists(s in S, exists (c in C0, trustedClients0[c] "
             "and s = connectedTo0[c] "
             "and next(trustedClients0[c]) = false "
             "and next(trustedServers[s]) = false));\n"))
    indent -= 2

def gen_end(sz, f):
    write(f, "end sys.impl;\n")

def write(f, s):
    global indent
    for i in range(0, indent):
        f.write("    ")
    f.write(s)
