#!/usr/bin/env python3
import os
import signal
import sys
import importlib
import subprocess
import time

CHECK = "./check_silent"
TIMEOUT = 0.0
RESULTS = "./results"
MODELS = "./models"

def run_tests(model, sizes):
    res = list(map(lambda sz: run_single_test(model, sz), sizes))
    with open(RESULTS+"/"+"results_"+model+".out", "w") as f:
        for e in res:
            f.write("\t".join(e)+"\n")

def run_single_test(model, size):
    gen = importlib.import_module(model + "_generator", package=None)
    (src, dst) = gen.get_src_dst(size)
    model_file = MODELS+"/"+model+"_"+str(size)+"_"+src+"_"+dst
    gen.generate(size, model_file)
    with open(model_file, 'r') as f:
        start = time.time()
        proc = subprocess.Popen([CHECK, "-src", src, "-dst", dst],
                stdin=f, stdout=subprocess.PIPE, preexec_fn=os.setsid)
        try:
            res = proc.communicate(timeout=TIMEOUT)[0]
            elapsed = time.time() - start
            res = [model, str(size), src, dst, str(elapsed)]\
                  + (res.decode("ascii").strip().split("="))[1].split(":")
        except subprocess.TimeoutExpired:
            os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
            res = [model, str(size), src, dst, "TIMEOUT("+str(TIMEOUT)+")"] + 6*[""]
        except:
            res = [model, str(size), src, dst, "EXIT"] + 6*[""]
    print(res)
    return res

# Usage ./tester.py model start end timeout
if __name__=='__main__':
    model = sys.argv[1]
    start = int(sys.argv[2])
    end = int(sys.argv[3]) + 1
    TIMEOUT = float(sys.argv[4])
    run_tests(model, range(start, end))
